#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pygame
import sys
import math

from stubbornchannels import *

class FarViz:
	def __init__(self):
		self.init = None
		self.screen = None
		self.screenorig = None
		self.surface = None
		self.changed = 1
		self.width = 400
		self.height = 400
		self.interrupted = 0
		self.cancelled = 0
		self.scheduler = None

		self.nodepositions = {}
		self.nodepositionsmoved = {}

	def dummy(self):
		n1 = node()
		n2 = node()

	def setup(self):
		if not self.init:
			pygame.init()
			pygame.display.set_caption("FarViz")
			self.interrupted = 0

		if not self.screen:
			self.screen = pygame.display.set_mode((self.width, self.height), pygame.DOUBLEBUF)
			#self.screen.fill((255, 240, 100))
			self.screen.fill((155, 140, 50))
			self.surface = None
			self.screenorig = pygame.Surface(self.screen.get_size())
			self.screenorig.blit(self.screen, (0, 0))

		if not self.surface:
			self.surface = pygame.Surface(self.screen.get_size())
			self.changed = 1

	def run(self):
		while not self.cancelled:
			self.setup()
			while not self.interrupted:
				self.simulation()
			self.screen = None

	def simulation(self):
		radiusx = self.screen.get_width() / 4
		radiusy = self.screen.get_height() / 4
		centerx = self.screen.get_width() / 2
		centery = self.screen.get_height() / 2

		if self.changed:
			self.changed = 0

			self.surface.blit(self.screenorig, (0, 0))

			font = pygame.font.SysFont("Vera Sans", 14)
			smallfont = pygame.font.SysFont("Vera Sans", 10)

			count = len(self.scheduler.nodes)
			self.nodepositions = {}

			i = 0
			for n in self.scheduler.nodes:
				degreex = math.sin(2.0 * math.pi * i / count)
				degreey = math.cos(2.0 * math.pi * i / count)
				x = int(degreex * radiusx) + centerx
				y = int(degreey * radiusy) + centery
				self.nodepositions[n] = (x, y)
				i += 1

			for n in self.nodepositionsmoved.keys():
				if self.nodepositions.has_key(n):
					self.nodepositions[n] = self.nodepositionsmoved[n]

			for c in self.scheduler.channels:
				if not c.sendnode:
					continue
				if c.backchannel:
					continue
				(xs, ys) = self.nodepositions[c.sendnode]
				(xr, yr) = self.nodepositions[c.receivenode]

				color = (0, 0, 0)
				if c.message:
					color = (255, 0, 0)
				if c.ack:
					color = (0, 255, 0)

				pygame.draw.aaline(self.surface, color, (xs, ys), (xr, yr), 0)

			for n in self.scheduler.nodes:
				r = n.receiver
				if r.host is not None:
					pass
				(x, y) = self.nodepositions[n]

				pygame.draw.circle(self.surface, (0, 0, 0), (x, y), 50)
				f = font.render(n.name, 1, (255, 255, 255))
				self.surface.blit(f, (x - f.get_width() / 2, y - 30))
				url = ""
				comment = "(local)"
				if r.url:
					url = r.url
					if r.islocal:
						comment = "(local-net)"
					else:
						comment = "(remote)"
				f = smallfont.render(url, 1, (150, 150, 150))
				self.surface.blit(f, (x - f.get_width() / 2, y - 10))
				f = smallfont.render(comment, 1, (150, 150, 150))
				self.surface.blit(f, (x - f.get_width() / 2, y))

			self.screen.blit(self.surface, (0, 0))

		pygame.event.pump()
		event = pygame.event.poll()
		if event.type == pygame.KEYDOWN:
			key = event.key
			if key == pygame.K_ESCAPE or pygame.event.peek(pygame.QUIT):
				self.cancelled = 1
				self.interrupted = 1
			if key == pygame.K_PLUS:
				self.width += 100
				self.height += 100
				self.interrupted = 1

		pygame.display.flip()

	def mousemove(self, x1, y1, x2, y2):
		print "==", x1, y1, x2, y2
		for n in self.nodepositions.keys():
			(x, y) = self.nodepositions[n]
			print "check", x, y
			if x1 > x - 10 and x1 < x + 10:
				if y1 > y - 10 and y1 < y + 10:
					x += x2 - x1
					y += y2 - y1
					self.nodepositionsmoved[n] = (x, y)
					self.changed = 1
					self.simulation()

print "[farviz] Launch!"

if __name__ == "__main__":
	fv = FarViz()
	fv.scheduler = scheduler
	fv.dummy()
	fv.run()

