# -*- coding: utf-8 -*-
#
# Klasse zur verteilten Programmausführung auf Unix-Ebene
#
# Projekt für den Großen Beleg
# 12.09.2005
# Josef Spillner <js177634@inf.tu-dresden.de>

import os
import sys
import thread

import common

class Remote:
	def __init__(self, host, port = None):
		self.host = host
		self.port = port
		self.prepared = 0
		self.dir = "__farnode"
		self.verbose = 0
		self.currentoutput = []
		self.partition = None

		if self.host.find("!") >= 0:
			(bangpath, self.host) = self.host.split("!")
			if bangpath.find(":") >= 0:
				(self.host, self.port) = bangpath.split(":")
			else:
				self.host = bangpath
		if self.host.find("/") >= 0:
			(self.host, self.partition) = self.host.split("/")
		if self.host.find(":") >= 0:
			(self.host, port) = self.host.split(":")

		self.lock = thread.allocate_lock()
		self.active = 1

	def debug(self, s):
		if self.verbose:
			print s

	def cancel(self):
		self.active = 0

	def islocal(self):
		localhosts = common.nodenames()
		# TODO: honour username?
		return (self.host in localhosts)

	def rexecute(self, cmd):
		if self.islocal():
			url = cmd
			if len(url) > 2:
				if url[0] == "'" and url[-1:] == "'":
					url = "sh -c " + cmd
		else:
			portarg = ""
			if self.port:
				portarg = "-p " + str(self.port) + " "
			url = "ssh " + portarg + self.host + " " + cmd
		self.debug("-- " + url)
		(fin, fout) = os.popen4(url)

		retout = []
		retoutnl = []
		self.active = 1
		while self.active:
			try:
				line = fout.readline()
			except:
				line = ""
			#print ">>> READ", "'" + line.strip() + "'"
			if line == "":
				self.active = 0
			else:
				retoutnl.append(line)
				retout.append(line.strip())

				#print "(thread::lock)"
				self.lock.acquire()
				self.currentoutput.append(line.strip())
				self.lock.release()
				#print "(thread::unlock)"

		try:
			ret = fout.close()
		except:
			ret = -9999

		# quick hack
		if len(retoutnl) == 1:
			e = retoutnl[0]
			if e.find("Command not found.\n") != -1:
				ret = -9999
			if e.find("command not found\n") != -1:
				ret = -9999
			if e.find("Connection refused\r\n") != -1:
				ret = -9999

		if ret:
			return (ret, retout)
		return (0, retout)

	def rrun(self, cmd):
		cmd = "'cd " + self.dir + " && " + cmd + "'"
		return self.rexecute(cmd)

	def rput(self, file):
		if not self.prepared:
			self.debug("-- mkdir first")
			(ret, out) = self.rexecute("mkdir -p " + self.dir)
			if not ret:
				self.prepared = 1
			else:
				return ret
		self.debug("-- put " + file)
		if self.islocal():
			return os.system("cp " + file + " " + self.dir)
		else:
			portarg = ""
			if self.port:
				portarg = "-P " + str(self.port) + " "
			return os.system("scp " + portarg + file + " " + self.host + ":" + self.dir)

	def rget(self, file):
		self.debug("-- get " + file)
		if self.islocal():
			return os.system("cp " + self.dir + "/" + file + " .")
		else:
			portarg = ""
			if self.port:
				portarg = "-P " + str(self.port) + " "
			return os.system("scp " + portarg + self.host + ":" + self.dir + "/" + file + " .")

	def setup(self):
		if self.islocal():
			self.debug("-- setup skipped (local node)")
			return 0
		self.debug("-- setup")
		home = os.getenv("HOME")
		prvkey = home + "/.ssh/identity"
		pubkey = prvkey + ".pub"
		try:
			x = os.stat(pubkey)
		except:
			self.debug("-- setup ssh keys first")
			ret = os.system("ssh-keygen -t dsa -f " + prvkey)
			if ret != 0:
				return ret

		self.debug("-- copy pubkey")
		redir = ""
		if self.verbose == 0:
			redir = " >/dev/null"
		if self.port:
			# FIXME: ssh-copy-id cannot handle ports!
			return 0
		return os.system("ssh-copy-id -i " + pubkey + " " + self.host + redir)

