#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Simulationsumgebung für Stubborn Channels.
#
# Projekt für den Großen Beleg
# 06.06.2005 - 04.07.2005
# Josef Spillner <js177634@inf.tu-dresden.de>

""" Import von Modulen """

import socket
import pickle
import time
import random
import math
import os
import sys

""" Globale Funktionen (von farlib/check.py) """

reset = "\033[0m"
green = "\033[92m"
darkgreen = "\033[32m"
yellow = "\033[93m"
red = "\033[91m"
cyan = "\033[94m"
blue = "\033[34m"
console = "\033[90m"
fat = "\033[1m"

__interrupthandler = None

def setinterrupthandler(h):
	global __interrupthandler

	__interrupthandler = h

__eventhandler = None

def seteventhandler(h):
	global __eventhandler

	__eventhandler = h

__tracehandler = None

def settracehandler(h):
	global __tracehandler

	__tracehandler = h

__dumbterminal = 0

def setdumbterminal():
	global __dumbterminal

	__dumbterminal = 1

def colour(attribute, s):
	global __dumbterminal

	if __dumbterminal == 1:
		return s
	s = attribute + s + reset
	return s

""" Globale Funktionen """

__debug = 0
__quiet = 0

def setverbose():
	global __debug

	__debug = 1

def debug(s):
	global __debug

	if __debug:
		print "# " + s

def setquiet():
	global __quiet

	__quiet = 1

def info(s):
	global __quiet

	if __quiet:
		return
	print colour(cyan, "INFO: " + s)

def appinfo(s):
	global __quiet

	if __quiet:
		return
	print colour(blue, "INFO: " + s)

def warning(s):
	global __quiet

	if __quiet:
		return
	print colour(yellow, "WARNING: " + s)

def error(s):
	global __quiet

	if __quiet:
		return
	print colour(red, "ERROR: " + s)

def fdstatus(s):
	global __quiet

	if __quiet:
		return
	now = scheduler.time()
	print colour(darkgreen, "FD: " + str(now) + " " + s)
	sys.stdout.flush()

def status(s):
	global __quiet

	if __quiet:
		return
	now = scheduler.time()
	print colour(green, "STATUS: " + str(now) + " " + s)
	sys.stdout.flush()

__counter = 0

def counter():
	global __counter

	__counter += 1
	return __counter

__portcounter = 30000

def portcounter():
	global __portcounter

	__portcounter += 1
	return __portcounter

__partition = None

def setpartition(partition):
	global __partition

	__partition = partition

__instances = 1

def setinstances(instances):
	global __instances

	__instances = instances

# Kopie von common.interfacenodes()
def interfacenodes():
	cmd = "LANG=C /sbin/ifconfig"
	cmd += "| grep \"inet addr\""
	cmd += "| cut -d \":\" -f 2"
	cmd += "| cut -d \" \" -f 1"
	(fin, fout) = os.popen2(cmd)

	ret = []
	while 1:
		addr = fout.readline()
		if addr == "":
			break
		addr = addr.strip()
		ret.append(addr)
	return ret

# Kopie von common.nodenames()
# erweitert um Partitionierung
def nodenames():
	global __partition

	tmp = []
	localhostnames = []

	(name, aliases, ips) = socket.gethostbyaddr("127.0.0.1")
	tmp.append(name)
	for alias in aliases:
		tmp.append(alias)
	for ip in ips:
		tmp.append(ip)
	tmp.append(socket.getfqdn())
	tmp.append(socket.gethostname())

	tmp += interfacenodes()

	for host in tmp:
		if not host in localhostnames:
			localhostnames.append(host)
			localhostnames.append(os.getenv("USER") + "@" + host)
			if __partition:
				host = host + "/" + __partition
				localhostnames.append(host)
				localhostnames.append(os.getenv("USER") + "@" + host)			
	return localhostnames

""" Klasse zur Repräsentation eines Kanals """

class Channel:
	def __init__(self, type, droprate = 0.0, droptimer = 0.0, buffersize = 0):
		subtype = None
		if type:
			if type.find(":") >= 0:
				(type, subtype) = type.split(":")
		if type not in ("reliable", "stubborn", "best-effort", "unreliable"):
			error("unknown type " + type + ", defaulting to stubborn")
			type = "stubborn"
		if type == "stubborn":
			if subtype == "tcp":
				error("stubborn channels must not use tcp")
		if type == "reliable":
			if subtype != "internal":
				error("only internal channels are reliable")

		self.type = type
		self.subtype = subtype

		self.receivenode = None
		self.sendnode = None
		self.message = None
		self.ack = None
		self.warned = 0

		if droprate < 0.0 or droprate > 1.0:
			error(self.id() + "drop rate probability out of range, reset")
			droprate = 0.0

		self.droprate = droprate
		self.droptimer = droptimer

		self.buffersize = buffersize

		self.backchannel = 0
		self.socket = None

		self.lastsocketusage = 0
		self.lastsocketmessage = None
		self.lastsocketack = None

		self.delayuntil = 0

	def connect(self, sendnode, receivenode):
		self.sendnode = sendnode
		self.receivenode = receivenode

	def send_message(self, message):
		debug("(channel) received message for sending: [" + str(message) + "]")
		if self.buffersize > 0:
			(header, msg) = message
			if len(msg) > self.buffersize:
				msg = msg[0:self.buffersize]
				message = (header, msg)
				warning(self.id() + "received large message, cut to buffer size")
		self.message = message
		#self.ack = None
		self.warned = 0
		self.process()
		status(self.id() + "sending message [" + str(message) + "]")

	def send_ack(self, ack):
		debug("(channel) received ack for sending: [" + str(ack) + "]")
		self.ack = ack
		self.process()
		status(self.id() + "sending acknowledgement [" + str(ack) + "]")

	def handle_message(self, message):
		status(self.id() + "got message [" + str(message) + "]")

	def done(self, withcheck = None):
		#print "------------------ done; discard", self.message, "in", self.id()
		if not self.message:
			# FIXME: shouldn't happen!?
			return
		#self.message = None
		if self.ack and withcheck:
			#(header_ack, msg_ack) = self.ack
			header_ack = withcheck
			(header_message, msg_message) = self.message
			if header_ack == header_message:
				self.message = None
			elif self.warned == 0:
				warning("(channel) ack read after sending new message")
				debug(" ack should match " + str(header_message))
				debug(" but is " + str(header_ack))
				self.warned = 1
		else:
			self.message = None

	def listen(self):
		# ..
		# FIXME: move from scheduler?
		pass

	def do_send(self, what):
		if not self.lastsocketusage < time.time() - 1.0:
			nosend = 0
			if what == "ack" and self.lastsocketack == self.ack:
				nosend = 1
			if what == "message" and self.lastsocketmessage == self.message:
				nosend = 1
			if nosend and self.socket:
				debug("not sending across network - gentlemen's timeout")
				return

		receivetuple = (self.receivenode.receiver.host, self.receivenode.receiver.port)
		sendurltuple = (self.sendnode.receiver.url, self.sendnode.receiver.port)
		debug("sending across network!")
		debug("destination: " + str(sendurltuple))

		if not self.socket:
			if self.subtype == "udp":
				self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
			elif self.subtype == "tcp":
				self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
				self.socket.connect(receivetuple)
			self.socket.setblocking(0)
		message = self.message
		if what == "ack":
			message = self.ack
		data = pickle.dumps((sendurltuple, message))

		bytes = -1
		if self.subtype == "udp":
			try:
				bytes = self.socket.sendto(data, 0, receivetuple)
			except KeyboardInterrupt:
				raise
			except: # gaierror?
				bytes = -1
		elif self.subtype == "tcp":
			try:
				bytes = self.socket.send(data)
			except KeyboardInterrupt:
				raise
			except: # gaierror?
				bytes = -1

		if bytes >= 0:
			debug("sent bytes: " + str(bytes))
		else:
			error("sending failed")

		self.lastsocketusage = time.time()
		if what == "message":
			self.lastsocketmessage = self.message
		elif what == "ack":
			self.lastsocketack = self.ack
		if self.type != "stubborn":
			self.done()

	def process(self):
		probability = 100 * self.droprate
		rand = random.randint(0, 100)
		if rand < probability:
			self.message = None
			# FIXME: is ack used at all still?
			self.ack = None
			#debug(self.id() + "ouch, dropped a message")
			print self.id() + "ouch, dropped a message" # FIXME: debug only?

	def id(self):
		s = "???"
		r = "???"
		if self.sendnode:
			s = self.sendnode.name
		if self.receivenode:
			r = self.receivenode.name
		return "<Channel <" + s + ">-<" + r + ">> "

""" Klasse zur Repräsentation eines Senders """

class Sender:
	def __init__(self, name = None, crash = 0.0, crashtimer = 0.0):
		self.name = "sender#???"
		if name is None:
			info(self.id() + "no name given, will generate one")
			name = "sender#" + str(counter())
		self.name = name

		if crash < 0.0 or crash > 1.0:
			error(self.id() + "crash probability out of range, reset")
			crash = 0.0
		self.crash = crash
		self.crashtimer = crashtimer
		self.crashed = 0

		self.times = 0
		self.pendingtimes = {}
		self.occupied = {}

		self.timeouts = {}
		self.slowmessages = {}
		self.fastmessages = {}

		self.slownodes = {}
		self.fastnodes = {}

		self.usequeues = 0
		self.queue = []
		self.queueack = None

		self.fdmodel = "ea" # "ps" or "ea"

	def generate_message(self, message, receiver):
		if self.crashed:
			error(self.id() + "crashed already")
			return
		channel = scheduler.connection(self.parentnode, receiver)
		if channel:
			if channel.type != "stubborn":
				if self.occupied.has_key(receiver):
					msgid = self.occupied[receiver]
					del self.occupied[receiver]
					del self.pendingtimes[msgid]
		if self.occupied.has_key(receiver):
			if channel.type == "stubborn":
				warning(self.id() + "ought not send new message, still waiting for ack from " + receiver.name)
				#return None
				if self.usequeues:
					self.queue.append((message, receiver))
					return None
		debug(self.id() + "go sleep and wait for ack")
		msgid = counter()
		self.occupied[receiver] = msgid
		self.pendingtimes[msgid] = scheduler.time()
		self.process()
		if self.usequeues:
			channel = scheduler.connection(self.parentnode, receiver)
			if channel and channel.buffersize > 0 and len(message) > channel.buffersize:
				retain = message[channel.buffersize:]
				message = message[0:channel.buffersize]
				self.queue.append((retain, receiver))
				warning(self.id() + "large message received, queued parts of it")
		debug(self.id() + "generated message " + str(msgid) + " for " + receiver.name)
		return (msgid, message)

	def checkqueue(self):
		if len(self.queue) > 0:
			warning(self.id() + "process next message from queue")
			self.queue.reverse()
			item = self.queue.pop()
			self.queue.reverse()
			#print "ITEM", item
			(message, receiver) = item
			message = self.generate_message(message, receiver)
			if message:
				channel = scheduler.connection(self.parentnode, receiver)
				if channel:
					channel.send_message(message)
					return 1
			else:
				self.queue.append(item)
		return 0

	def read_ack(self, receiver, ack):
		debug(self.id() + "read ack from " + receiver.name)
		if self.crashed:
			error(self.id() + "crashed already")
			return 1
		if not self.occupied.has_key(receiver):
			debug(self.id() + "won't read ack, not waiting for any")
			return 1
		(msgid, message) = ack
		if not self.pendingtimes.has_key(msgid):
			debug("received ack for message which was never sent")
			return 1
		status(self.id() + "received message (ack) " + str(msgid) + " from " + receiver.name)
		receivetimer = scheduler.time()
		sendtimer = self.pendingtimes[msgid]
		del self.pendingtimes[msgid]
		del self.occupied[receiver]
		if receivetimer < sendtimer:
			error(self.id() + "time reset detected")
		else:
			if self.fdmodel == "ea":
				diff = receivetimer - sendtimer
				debug(self.id() + "time calculation: " + str(diff))
				if not self.timeouts.has_key(receiver):
					self.timeouts[receiver] = 1.0
					self.slowmessages[receiver] = 0
					self.fastmessages[receiver] = 0
				if diff > self.timeouts[receiver]:
					self.slowmessages[receiver] += 1
					self.fastmessages[receiver] = 0
					debug(self.id() + "message is slow!")
				else:
					self.fastmessages[receiver] += 1
					debug(self.id() + "message is fast!")
			# always un-suspect node on arrival of ack
			if self.fdmodel == "ea":
				self.fastnodes[receiver] = 1
				if self.slownodes.has_key(receiver):
					del self.slownodes[receiver]
					if self.parentnode.fdcallbackmethod:
						method = self.parentnode.fdcallbackmethod
						method.__call__(receiver, 1)
			self.times += 1
			if self.fdmodel == "ea":
				debug(self.id() + str(self.fastmessages[receiver]) + " fast messages")
				debug(self.id() + str(self.slowmessages[receiver]) + " slow messages")
				newtimeout = (1 + self.fastmessages[receiver]) * (1 + math.log(1 + self.slowmessages[receiver]))
				self.timeouts[receiver] = newtimeout
				debug(self.id() + "new timeout: " + str(self.timeouts[receiver]))
				fm = str(self.fastmessages[receiver])
				sm = str(self.slowmessages[receiver])
				to = str(self.timeouts[receiver])
				di = str(diff)
				fdstatus(self.id() + "EA-FD: " + fm + " fast, " + sm + " slow, diff = " + di + ", timeout = " + to)
		self.process()
		if self.usequeues:
			if self.checkqueue():
				return 0
		return 1

	def process(self):
		probability = 100 * self.crash
		rand = random.randint(0, 100)
		if rand < probability:
			self.crashed = 1
			debug(self.id() + "ouch, crashed")

	def id(self):
		return "(sender " + self.name + ") "

""" Klasse zur Repräsentation eines Empfängers """

class Receiver:
	def __init__(self, name = None, url = None, port = None, crash = 0.0, crashtimer = 0.0):
		self.name = "receiver#???"
		if name is None:
			info(self.id() + "no name given, will generate one")
			name = "receiver#" + str(counter())
		self.name = name
		self.url = url
		self.host = url
		self.port = None
		self.proxyhost = None
		self.proxyport = None

		partition = None
		user = None

		if url:
			if self.host.find("!") >= 0:
				(bangpath, self.host) = self.host.split("!")
				if bangpath.find(":") >= 0:
					(self.proxyhost, self.proxyport) = bangpath.split(":")
				else:
					self.proxyhost = bangpath
			if self.host.find("@") >= 0:
				(user, self.host) = self.host.split("@")
			if self.host.find("/") >= 0:
				(self.host, partition) = self.host.split("/")
			if self.host.find(":") >= 0:
				if port is not None:
					warning(self.id() + "url port overwrites explicit port")
				(self.host, port) = self.host.split(":")
				port = int(port)

			if port is None:
				warning(self.id() + "no port given, will generate one")
				port = portcounter()

		self.port = port

		# FIXME: only in sender for now, move to node
		#if crash < 0.0 or crash > 1.0:
		#	error(self.id() + "crash probability out of range, reset")
		#	crash = 0.0
		#self.crash = crash
		#self.crashtimer = crashtimer
		#self.crashed = 0

		self.sockin = None
		self.islocal = 0
		if self.url:
			noporturl = self.host
			if user:
				noporturl = user + "@" + noporturl
			if partition:
				noporturl = noporturl + "/" + partition
			if noporturl in nodenames():
				self.islocal = 1
		else:
			self.islocal = 1

		self.delay = 0
		self.delayuntil = 0
		self.delaytraces = {}

		self.lastalives = {}
		self.lastsentalives = {}
		self.alivefrequency = 20.0

		self.useblocking = 0

	def generate_ack_postroutine(self, sender, message):
		pass

	def generate_ack(self, sender, message):
		debug(self.id() + "received message from " + sender.name + ", send ack")
		#rand = random.randint(0, 100)
		#if rand > 20:
		#	debug("(receiver) not yet ready for ack")
		#	return None

		if self.parentnode.sender.occupied.has_key(sender):
			channel = scheduler.connection(sender, self.parentnode)
			msgid = self.parentnode.sender.occupied[sender]
			sendtimer = self.parentnode.sender.pendingtimes[msgid]
			unblock = 0
			if scheduler.time() > sendtimer + self.parentnode.timeout(sender):
				unblock = 1
			if channel and channel.type == "stubborn" and not unblock:
				warning(self.id() + "ought not accept this message, still waiting for ack from " + sender.name)
				if self.useblocking:
					return None

		if self.delayuntil == 0:
			rand2 = 0
			if self.delay > 0:
				rand2 = random.randint(0, self.delay)
			if self.delay < 0:
				rand2 = random.random() * 100
			if rand2 > 0:
				self.delayuntil = scheduler.time() + rand2
		delayuntil = self.delayuntil
		channel = scheduler.connection(sender, self.parentnode)
		if channel.delayuntil:
			delayuntil = channel.delayuntil
		if delayuntil == 0:
			if self.delaytraces.has_key(sender):
				#traces = self.delaytraces[sender]
				nextdelay = self.delaytraces[sender][0]
				self.delaytraces[sender] = self.delaytraces[sender][1:]
				if len(self.delaytraces[sender]) == 0:
					del self.delaytraces[sender]
				delayuntil = scheduler.time() + nextdelay
				channel.delayuntil = delayuntil
				appinfo("Trace " + str(nextdelay) + " on " + sender.name + " in " + self.parentnode.name)
		if delayuntil > scheduler.time():
			#print "TIME: it's", scheduler.time(), "wait until", self.delayuntil
			return None
		self.delayuntil = 0
		channel.delayuntil = 0

		(header, message) = message

		# always un-suspect node on arrival of ALIVE message
		if message == "ALIVE":
			if self.parentnode.sender.fdmodel == "ps":
				self.parentnode.sender.fastnodes[sender] = 1
				if self.parentnode.sender.slownodes.has_key(sender):
					del self.parentnode.sender.slownodes[sender]
					fdstatus("PS-FD of " + self.parentnode.name + " unsuspecting " + sender.name)
					if not self.parentnode.sender.timeouts.has_key(sender):
						self.parentnode.sender.timeouts[sender] = 1.0
					self.parentnode.sender.timeouts[sender] += 5.0
					fdstatus("(" + self.parentnode.name + ") PS-FD: timeout = " + str(self.parentnode.sender.timeouts[sender]))
					if self.parentnode.fdcallbackmethod:
						method = self.parentnode.fdcallbackmethod
						method.__call__(sender, 1)
				self.lastalives[sender] = scheduler.time()
				if self.lastsentalives.has_key(sender):
					del self.lastsentalives[sender]
				#return None

		ack = (header, "ACK:")
		return ack

	def set_delay(self, delay):
		self.delay = delay
		self.delayuntil = 0

	def id(self):
		return "(receiver " + self.name + ") "

""" Klasse zur Repräsentation eines Prozesses """

class Node:
	def __init__(self, name = None, url = None, port = None, sender = None, receiver = None):
		self.name = "node#???"
		if name is None:
			info(self.id() + "no name given, will generate one")
			name = "node#" + str(counter())
		self.name = name
		self.url = url

		self.sender = sender
		if not sender:
			self.sender = Sender(name + ":s")
		self.sender.parentnode = self
		self.receiver = receiver
		if not receiver:
			self.receiver = Receiver(name + ":r", url, port)
		else:
			if not self.url:
				self.url = receiver.url
		self.receiver.parentnode = self

		self.callbacktimer = None
		self.callbackmethod = None
		self.fdcallbackmethod = None

		self.isidle = 0

	def callback(self, timer, callback):
		if timer:
			self.callbacktimer = scheduler.time() + timer
		else:
			self.callbacktimer = None
		self.callbackmethod = callback

	def fdcallback(self, callback):
		self.fdcallbackmethod = callback

	def idle(self):
		info(self.id() + "go idle")
		self.isidle = 1

	def id(self):
		return "(node " + self.name + ") "

	def timeout(self, n):
		t = 1.0
		if self.sender.timeouts.has_key(n):
			t = self.sender.timeouts[n]
		return t

	def __getstate__(self):
		odict = self.__dict__.copy()
		if odict.has_key("callbackmethod"):
			del odict["callbackmethod"]
		if odict.has_key("fdcallbackmethod"):
			del odict["fdcallbackmethod"]
		return odict

""" Klasse zur Repräsentation der Objektverwaltung """

class Scheduler:
	def __init__(self):
		self.channels = []
		self.nodes = []
		self.connections = {}

		self.trivialqueue = []

		self.cancelled = 0
		self.stopped = 0

		self.time_offset = time.time()

	def register(self, object):
		if isinstance(object, Channel):
			if object in self.channels:
				error("Channel " + str(object) + " already registered")
			else:
				debug("channel registered with scheduler")
				self.channels.append(object)
		elif isinstance(object, Node):
			if object in self.nodes:
				error("Node " + str(object) + " already registered")
			else:
				debug("node registered with sender")
				self.nodes.append(object)
		else:
			error("registering unknown object " + str(object))

	def autoconnect(self, type, droprate = 0.0, droptimer = 0.0):
		for n in self.nodes:
			for m in self.nodes:
				if n != m:
					if self.connections.has_key(str(m) + str(n)):
						continue
					if self.connections.has_key(str(n) + str(m)):
						continue
					info("AUTOCONNECT " + n.name + " with " + m.name)
					c = channel(type, droprate, droptimer)
					connect(m, n, c)

	def connect(self, sendnode, receivenode, channel):
		if sendnode == receivenode:
			error("channel not possible between two times the same node")
			return
		if not channel.subtype:
			if sendnode.receiver.url:
				channel.subtype = "udp"
			else:
				channel.subtype = "internal"
			info("setting channel subtype to " + channel.subtype)
		if channel.subtype:
			if channel.subtype in ("tcp", "udp", "tcp-r"):
				if not (sendnode.receiver.url and receivenode.receiver.url):
					error("udp/tcp channel between local nodes")
			if channel.subtype == "internal":
				if sendnode.receiver.url or receivenode.receiver.url:
					error("internal channel between remote nodes")
		if channel.sendnode:
			warning("disconnecting channel from sender " + channel.sendnode.name)
		if channel.receivenode:
			warning("disconnecting channel from receiver " + channel.receivenode.name)
		self.connections[str(sendnode) + str(receivenode)] = channel
		channel.connect(sendnode, receivenode)

		# full duplex connections
		if not self.connections.has_key(str(receivenode) + str(sendnode)):
			c = Channel(channel.type, channel.droprate, channel.droptimer)
			c.subtype = channel.subtype
			c.backchannel = 1
			self.register(c)
			self.connect(receivenode, sendnode, c)

	def connection(self, sendnode, receivenode):
		c = None
		key = str(sendnode) + str(receivenode)
		if self.connections.has_key(key):
			c = self.connections[key]
		else:
			error("no channel found between " + sendnode.name + " and " + receivenode.name)
		return c

	def findnode(self, addr):
		(url, port) = addr
		for n in self.nodes:
			r = n.receiver
			if r.url == url and r.port == port:
				return n
		return None

	def getnode(self, name):
		for n in self.nodes:
			if n.name == name:
				return n
		return None

	def information(self):
		info("# Scheduler information:")

		channelcount = 0
		for c in self.channels:
			if c.backchannel:
				continue
			channelcount += 1

		info("# " + str(channelcount) + " channels")
		for c in self.channels:
			if c.backchannel:
				continue
			info("# " + c.id() + "(" + c.type + "/" + c.subtype + ")")
			if not c.sendnode:
				info("#  " + "(unconnected)")
			else:
				tmp = "#  " + "(" + c.sendnode.name + "->" + c.receivenode.name + ")" + ": "
				if c.message:
					info(tmp + "active")
					#print "active (message =", c.message, ")"
				else:
					if c.sendnode.receiver.islocal:
						info(tmp + "inactive")
					else:
						info(tmp + "(no information)")

		info("# " + str(len(self.nodes)) + " nodes")
		for n in self.nodes:
			s = n.sender
			r = n.receiver

			tmp = "# " + n.name + " [FD: " + n.sender.fdmodel + "] "
			if n.isidle:
				info(tmp + "(idle)")
			else:
				info(tmp)
			tmp = "#  [s] " + s.name + ": "
			tmp += str(len(s.pendingtimes)) + " messages pending "
			tmp += str(s.times) + " done"
			info(tmp)
			for m in s.occupied.keys():
				info("#   message to " + m.name)

			tmp = "#  [r] " + r.name + ": "
			if r.url:
				if r.islocal:
					info(tmp + "local-net " + r.url + ":" + str(r.port))
				else:
					info(tmp + "remote " + r.url + ":" + str(r.port))
			else:
				info(tmp + "local")

			if r.islocal:
				info("#  [suspected nodes]")
				for m in s.slownodes:
					info("#   (suspected) " + m.name)
				for m in s.fastnodes:
					if m == n:
						info("#   (self) " + m.name)
					else:
						info("#   (ok) " + m.name)

	def cancel(self):
		self.cancelled = 1

	def stop(self):
		self.stopped = 1

	def unstop(self):
		self.stopped = 0

	def work(self):
		if self.stopped:
			return

		for m in self.trivialqueue:
			(node, message) = m
			node.receiver.generate_ack_postroutine(node, (-1, message))
		self.trivialqueue = []

		debug("work (receive from network)...")
		for n in self.nodes:
			r = n.receiver
			r.activated = 0

			if r.url and r.islocal:
				debug("recv on " + r.name)
				if not r.sockin:
					subtype = "udp" # FIXME
					if subtype == "udp":
						r.sockin = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
					elif subtype == "tcp":
						r.sockin = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
					r.sockin.setblocking(0)
					r.sockin.bind((r.host, r.port))
					debug("binding to " + str(r.url) + ":" + str(r.port))

				try:
					subtype = "udp" # FIXME
					if subtype == "udp":
						(data, addr) = r.sockin.recvfrom(1024)
					elif subtype == "tcp":
						data = r.sockin.recv(1024)
				except:
					#warning("nothing received")
					debug("nothing received")
					continue

				r.activated = 1

				#print "GOT", data, addr
				#c.ack = c.listen() # bleh...
				(addr, message) = pickle.loads(data)

				originnode = scheduler.findnode(addr)
				if not originnode:
					debug("sender is unknown, dismiss")
					continue

				(header, content) = message

				if isinstance(content, str) and content.find("ACK:") == 0:
					debug("received ack: " + content)
					# and now?
					ret = n.sender.read_ack(originnode, message)
					if ret == 1:
						channel = scheduler.connection(n, originnode)
						if channel:
							channel.done(header)
						else:
							warning("ack will not stop message")
					# TODO: acks will be sent forever?
				else:
					debug("received message: " + str(content))

					channel = scheduler.connection(originnode, n)
					if channel:
						channel.handle_message(message)
					else:
						warning("received over anonymous channel")

					ack = r.generate_ack(originnode, message)
					if ack:
						debug("prepare ack for sending")
						channel = scheduler.connection(n, originnode)
						if channel:
							if channel.type == "stubborn":
								channel.send_ack(ack)
								channel.do_send("ack")
							#else:
							#	channel.done()
						r.generate_ack_postroutine(originnode, message)
					else:
						debug("ack delayed")

		debug("work (schedule)...")
		for c in self.channels:
			c.sendnode.receiver.activated = 0
			c.receivenode.receiver.activated = 0

		for c in self.channels:
			if c.ack:
				if not c.receivenode.receiver.url:
					# magic!
					debug("channel needs sending ack")
					c.sendnode.receiver.activated = 1
					ret = c.sendnode.sender.read_ack(c.receivenode, c.ack)
					if ret == 1:
						(header, ack) = c.ack
						c.done(header)
				else:
					c.do_send("ack")
			if c.message:
				#print "--------- message", c.message, "in", c.id()
				debug("channel needs sending message")
				if not c.receivenode.receiver.url:
					# magic!
					c.receivenode.receiver.activated = 1
					ack = c.receivenode.receiver.generate_ack(c.sendnode, c.message)
					if ack:
						c.handle_message(c.message)
						debug("send local ack")
						message = c.message
						if c.type == "stubborn":
							c.send_ack(ack)
						else:
							c.done()
						c.receivenode.receiver.generate_ack_postroutine(c.sendnode, message)
					else:
						debug("ack delayed")
				else:
					c.do_send("message")

	def time(self):
		return time.time() - self.time_offset

""" API-Funktionen für Simulationsskripte """

def channel(type, droprate = 0.0, droptimer = 0.0, buffersize = 0):
	debug(": new channel")
	c = Channel(type, droprate, droptimer, buffersize)
	scheduler.register(c)
	return c

class MultiNode:
	def __init__(self):
		self.objects = []
		self.currentfunc = None
		self.hierarchy = []

	def __getattr__(self, x):
		if x != "hierarchy":
			self.currentfunc = x

		p = self.objects[0]
		for h in self.hierarchy:
			p = p.__dict__[h]

		if p.__dict__.has_key(x):
			if type(p.__dict__[x]) == type(self): # instance
				self.hierarchy.append(x)
				return self
			else:
				self.hierarchy = []
				return p.__dict__[x]

		return self.proxy

	def __setattr__(self, x, y):
		reserved = ["objects", "currentfunc", "hierarchy"]
		if x in reserved:
			self.__dict__[x] = y
			return

		orighierarchy = self.hierarchy
		self.hierarchy = []
		for o in self.objects:
			p = o
			for h in orighierarchy:
				p = p.__dict__[h]
			p.__dict__[x] = y

	def proxy(self, *args):
		for o in self.objects:
			p = o
			for h in self.hierarchy:
				p = p.__dict__[h]
			func = p.__class__.__dict__[self.currentfunc]
			argline = ""
			i = 0
			for a in args:
				argline += ", args[" + str(i) + "]"
				i += 1
			eval("func.__call__(p" + argline + ")")
		self.hierarchy = []

def node(name = None, url = None, port = None, sender = None, receiver = None):
	global __instances

	debug(": new node")
	n = Node(name, url, port, sender, receiver)
	scheduler.register(n)

	if __instances > 1:
		warning("Cloning node " + n.name + " " + str(__instances) + "-fold")
		sclone = pickle.dumps(sender)
		rclone = pickle.dumps(receiver)
		m = MultiNode()
		m.objects.append(n)
		for i in range(__instances - 1):
			sender = pickle.loads(sclone)
			receiver = pickle.loads(rclone)
			n = Node(None, url, port, sender, receiver)
			scheduler.register(n)
			m.objects.append(n)
		return m
	return n

def connect(sendnode, receivenode, channel):
	debug(": connect called")
	scheduler.connect(sendnode, receivenode, channel)

def autoconnect(type, droprate = 0.0, droptimer = 0.0):
	debug(":autoconnect called")
	scheduler.autoconnect(type, droprate, droptimer)

def send(sendnode, receivenode, message):
	debug(": send called")

	if not sendnode.receiver.islocal:
		debug("not on me but on " + sendnode.name)
		return

	if sendnode == receivenode:
		debug("trivial submission to self")
		scheduler.trivialqueue.append((sendnode, message))
		return

	channel = scheduler.connection(sendnode, receivenode)
	if channel:
		message = sendnode.sender.generate_message(message, receivenode)
		if message:
			channel.send_message(message)

def schedule(steps = None, selectset = None, timeout = None):
	global __interrupthandler
	global __tracehandler

	if __tracehandler:
		for c in scheduler.channels:
			traces = __tracehandler.__call__(c.sendnode, c.receivenode)
			if traces:
				c.receivenode.receiver.delaytraces[c.sendnode] = traces
				appinfo(c.id() + "trace loaded (" + str(len(traces)) + " items)")
		__tracehandler = None

	scheduler.cancelled = 0

	lastinterrupt = None
	while not scheduler.cancelled:
		try:
			schedule_internal(steps, selectset, timeout)
		except KeyboardInterrupt:
			if lastinterrupt:
				if time.time() < lastinterrupt + 0.5:
					raise KeyboardInterrupt
			lastinterrupt = time.time()
			appinfo("Internal keyboard interrupt handler!")
			if __interrupthandler:
				__interrupthandler.__call__()

def schedule_internal(steps = None, selectset = None, timeout = None):
	global __eventhandler

	debug(": schedule")
	if steps:
		for i in range(steps):
			scheduler.work()
		scheduler.cancel()
	elif timeout:
		entrytime = scheduler.time()
		while scheduler.time() < entrytime + timeout:
			scheduler.work()
			if scheduler.cancelled:
				break
		scheduler.cancel()
	else:
		terminator = 2
		while not scheduler.cancelled:
			scheduler.work()
			if scheduler.stopped:
				continue

			if __eventhandler:
				__eventhandler.__call__()

			allidle = 1
			for n in scheduler.nodes:
				#if n.isidle == 0 and n.receiver.islocal:
				if n.isidle == 0:
					allidle = 0
			if allidle:
				terminator -= 1
				if terminator == 0:
					scheduler.cancel()
					return

			if selectset:
				for n in selectset:
					if n.receiver.activated:
						debug("ACTIVATION: " + n.receiver.name)
						scheduler.cancel()
						return
			for n in scheduler.nodes:
				if n.callbackmethod:
					if n.receiver.islocal:
						if scheduler.time() > n.callbacktimer:
							info("CALLBACK: " + n.name)
							method = n.callbackmethod
							n.callbacktimer = None
							n.callbackmethod = None
							# FIXME: workaround for python 2.3
							(major, minor, micro, release, serial) = sys.version_info
							if minor == 3:
								method.__call__(n.receiver)
							else:
								method.__call__()
				suspected = []
				for msgid in n.sender.pendingtimes:
					sendtimer = n.sender.pendingtimes[msgid]
					for receiver in n.sender.occupied:
						if n.sender.slownodes.has_key(receiver):
							continue
						if n.sender.occupied[receiver] == msgid:
							if scheduler.time() > sendtimer + n.timeout(receiver):
								if n.sender.fdmodel == "ea":
									fdstatus("EA-FD of " + n.name + " suspecting " + receiver.name)
									suspected.append(receiver)

				for receiver in n.receiver.lastsentalives:
					if n.sender.slownodes.has_key(receiver):
						continue
					if scheduler.time() > n.receiver.lastsentalives[receiver] + n.timeout(receiver):
						if n.sender.fdmodel == "ps":
							fdstatus("PS-FD of " + n.name + " suspecting " + receiver.name)
							suspected.append(receiver)
							n.receiver.lastalives[receiver] = scheduler.time()

				if n.sender.fdmodel == "ps":
					if not n.receiver.islocal:
						continue
					for receiver in scheduler.nodes:
						if receiver == n:
							continue
						if not n.receiver.lastalives.has_key(receiver):
							n.receiver.lastalives[receiver] = 0
						if n.receiver.lastsentalives.has_key(receiver):
							if not n.receiver.lastsentalives[receiver] < scheduler.time() - (n.receiver.alivefrequency + n.timeout(receiver)):
								continue
						if n.receiver.lastalives[receiver] < scheduler.time() - n.receiver.alivefrequency:
							fdstatus("PS-FD: send ALIVE from " + n.name + " to " + receiver.name)
							message = "ALIVE"
							channel = scheduler.connection(n, receiver)
							if channel:
								tmp = n.sender.usequeues
								n.sender.usequeues = 1
								message = n.sender.generate_message(message, receiver)
								n.sender.usequeues = tmp
								if message and channel:
									channel.send_message(message)
								n.receiver.lastsentalives[receiver] = scheduler.time()

				for receiver in suspected:
					if n.sender.fastnodes.has_key(receiver):
						del n.sender.fastnodes[receiver]
					n.sender.slownodes[receiver] = 1

					#if n.sender.fdmodel == "ea":
					#	if n.sender.occupied.has_key(receiver):
					#		msgid = n.sender.occupied[receiver]
					#		del n.sender.occupied[receiver]
					#		del n.sender.pendingtimes[msgid]

					if n.sender.fdmodel == "ps":
						if n.receiver.lastsentalives.has_key(receiver):
							del n.receiver.lastsentalives[receiver]

					if n.fdcallbackmethod:
						method = n.fdcallbackmethod
						method.__call__(receiver, 0)

""" Hauptprogramm """

#random.seed()

scheduler = Scheduler()

