#!/usr/bin/env python

import re
import sys

# RECEIVER

#STATUS: 49.7385451794 <Channel <bizarre>-<se01>> got message [(4, ('prepare', 0, 'INSERT INTO account (moneytransfer) VALUES (0)'))]
#STATUS: 49.7389280796 <Channel <se01>-<bizarre>> sending acknowledgement [(4, 'ACK:')]

# SENDER

#STATUS: 49.7394421101 <Channel <se01>-<bizarre>> sending message [(4, ('p-ack', 0))]
#STATUS: 49.8886220455 (sender se01:s) received message (ack) 4 from bizarre

def parsedata(script, filter = None):
	try:
		f = file(script)
		d = f.readlines()
		f.close()
	except:
		return None

	ansi = re.compile("\x1b..{1,2}\x6d")

	r_input = re.compile("STATUS: (\d+\.\d+) <Channel <(\S+)>-<(\S+)>> got message \[.*\]")

	r_output = re.compile("STATUS: (\d+\.\d+) <Channel <(\S+)>-<(\S+)>> sending message \[\((\d+), .*\]")
	r_ack = re.compile("STATUS: (\d+\.\d+) \(\S+ \S+\) received message \(ack\) (\d+) from (\S+)")

	inputs = []
	outputs = []
	acks = []
	roundtrips = []
	ackset = {}

	for line in d:
		line = line.strip()
		line = ansi.sub("", line, 2)
		#print line

		m_input = r_input.match(line)
		if m_input:
			x_time = float(m_input.group(1))
			x_sender = m_input.group(2)
			#print "INPUT at", x_time, "from", x_sender
			inputs.append(x_time)

		m_output = r_output.match(line)
		if m_output:
			x_time = float(m_output.group(1))
			x_receiver = m_output.group(3)
			x_header = int(m_output.group(4))
			#print "OUTPUT at", x_time, "to", x_receiver
			outputs.append(x_time)

			ackset[x_header] = x_time

		m_ack = r_ack.match(line)
		if m_ack:
			x_time = float(m_ack.group(1))
			x_header = int(m_ack.group(2))
			x_sender = m_ack.group(3)
			#print "ACK at", x_time, "for", x_header, "from", x_sender
			acks.append(x_time)

			if ackset.has_key(x_header):
				diff = x_time - ackset[x_header]
				if not filter or x_sender == filter:
					roundtrips.append(diff)
				del ackset[x_header]

	print "**", len(inputs), "input messages found"
	print "**", len(outputs), "output messages found"
	print "**", len(acks), "ack messages found"
	print "**", len(roundtrips), "roundtrips calculated"

	data = (inputs, outputs, acks, roundtrips)

	return data

def xround(val):
	mult = 1000 # 3 digits
	val = float(int(val * mult))
	return val / mult

def overview_stats(valuelist, percentiles = 10, sections = 10, dodiff = True):
	diffsum = 0
	difflist = []
	old = valuelist[0]
	for i in range(1, len(valuelist)):
		if dodiff:
			diff = valuelist[i] - old
		else:
			diff = valuelist[i]
		diffsum += diff
		difflist.append(diff)
		old = valuelist[i]
	ordereddifflist = difflist[:]
	ordereddifflist.sort()
	print "Mean average time", xround(diffsum / len(valuelist))

	print "Percentiles:"
	a = len(ordereddifflist) / percentiles
	for i in range(percentiles):
		lower = i * a
		upper = (i + 1) * a
		slist = ordereddifflist[lower:upper]
		sdiffsum = 0
		for diff in slist:
			sdiffsum += diff
		counter = "(" + str("%2i") % (i + 1) + "/" + str(percentiles) + ")"
		print counter + " value percentile", xround(sdiffsum / len(slist))

	print "Time sections:"
	a = len(difflist) / sections
	for i in range(sections):
		lower = i * a
		upper = (i + 1) * a
		slist = difflist[lower:upper]
		sdiffsum = 0
		for diff in slist:
			sdiffsum += diff
		counter = "(" + str("%2i") % (i + 1) + "/" + str(sections) + ")"
		print counter + " time section", xround(sdiffsum / len(slist))

def overview(data):
	print "Received messages:"
	overview_stats(data[0])
	print "Sent messages:"
	overview_stats(data[1])
	print "Received acks:"
	overview_stats(data[2])
	print "Send/receive roundtrips:"
	overview_stats(data[3], 10, 10, False)

def injecttraces(diffs):
	for diff in diffs:
		print diff

if __name__ == "__main__":
	if len(sys.argv) != 2 and len(sys.argv) != 3:
		print "Error: no statistics file specified"
		sys.exit(1)
	script = sys.argv[1]
	filter = None
	if len(sys.argv) == 3:
		filter = sys.argv[2]

	data = parsedata(script, filter)
	if not data:
		print "Error: opening statistics file " + script + " failed"
		sys.exit(1)

	#overview(data)

	if filter:
		print filter + "->???"
	injecttraces(data[3])

