#!/usr/bin/env python

import pygame
import sys
import re

def display(data):
	pygame.init()
	pygame.display.set_caption("Simulation Statistics")

	screen = pygame.display.set_mode((800, 600), pygame.DOUBLEBUF)

	background = pygame.Surface((screen.get_width(), screen.get_height()))
	background.fill((0, 0, 100))
	screen.blit(background, (0, 0))

	font = pygame.font.SysFont("Vera Sans", 10)

	factor = 8
	xoff = 100
	yoff = 400

	inittime = 0
	for d in data:
		(x_time, x_diff, x_timeout) = d
		if not inittime:
			inittime = x_time
		x = int(xoff + (x_time - x_diff - inittime) * factor)
		twidth = int(x_timeout * factor)
		height = twidth

		pygame.draw.rect(screen, (0, 0, 70), ((x, yoff - height), (twidth, height)))

	first = 1
	for d in data:
		(x_time, x_diff, x_timeout) = d
		x = int(xoff + (x_time - x_diff - inittime) * factor)
		width = int(x_diff * factor)
		twidth = int(x_timeout * factor)
		height = width

		for i in range(width):
			colour = (0, 255, 0)
			if i > twidth:
				colour = (255, 0, 0)
			if i >= twidth - 3 and i <= twidth + 3:
				d = twidth + 3 - i
				colour = (255 - d * 40, d * 40, 0)
			h = height * (float(i) / width)
			pygame.draw.rect(screen, colour, ((x + i, yoff - h), (1, h)))

		x_time_r = int(x_time * 100) / 100.0
		x_diff_r = int(x_diff * 100) / 100.0
		x_timeout_r = int(x_timeout * 100) / 100.0

		if first:
			f = font.render("Response event:", 1, (255, 255, 255))
			screen.blit(f, (x, yoff + 30))

		f = font.render(str(x_time_r), 1, (255, 255, 255))
		screen.blit(f, (x, yoff + 40))

		if first:
			f = font.render("Roundtrip time:", 1, (255, 255, 255))
			screen.blit(f, (x, yoff + 60))

		f = font.render(str(x_diff_r), 1, (255, 255, 255))
		screen.blit(f, (x, yoff + 70))

		if first:
			f = font.render("Timeout at:", 1, (255, 255, 255))
			screen.blit(f, (x, yoff + 90))

		f = font.render(str(x_timeout_r), 1, (255, 255, 255))
		screen.blit(f, (x, yoff + 100))

		if first:
			first = 0

	while 1:
		event = pygame.event.wait()
		if event.type == pygame.KEYDOWN:
			key = event.key
			if key == pygame.K_ESCAPE or pygame.event.peek(pygame.QUIT):
				break

		pygame.display.flip()

def parsedata(script):
	try:
		f = file(script)
		d = f.readlines()
		f.close()
	except:
		return None

	r = re.compile("FD: (\d+\.\d+) \(\S+ \S+\) EA-FD: (\d+) fast, (\d+) slow, diff = (\d+\.\d+), timeout = (\d+\.\d+)")

	data = []
	for line in d:
		line = line.strip()
		if line == "" or line[0] == "#":
			continue

		m = r.match(line)
		if m:
			x_time = float(m.group(1))
			x_fast = int(m.group(2))
			x_slow = int(m.group(3))
			x_diff = float(m.group(4))
			x_timeout = float(m.group(5))
			data.append((x_time, x_diff, x_timeout))

	return data

if __name__ == "__main__":
	if len(sys.argv) != 2:
		print "Error: no statistics file specified"
		sys.exit(1)
	script = sys.argv[1]

	data = parsedata(script)
	if not data:
		print "Error: opening statistics file " + script + " failed"
		sys.exit(1)

	display(data)

